{# Displays links to the top-level TOCtree elements, in the header navbar. #}
<nav class="navbar-nav">
  <p class="sidebar-header-items__title"
     role="heading"
     aria-level="1"
     aria-label="{{ _('') }}">
    {{ _("") }}
  </p>
  <ul class="bd-navbar-elements navbar-nav">
    {{ generate_header_nav_html(n_links_before_dropdown=20, dropdown_text=theme_header_dropdown_text) }}
  </ul>
</nav>